## 0. check environment

const required_envs = [
    "EDLS_GITHUB_TOKEN"
    "CI_API_V4_URL"
    "CI_PROJECT_ID"
]
@assert all(haskey(ENV, var) for var ∈ required_envs) "Environment set up incorrectly. Make sure all of the following variables are set:\n\t"*join(required_envs, "\n\t")

const TESTING = false
const REG_URL_OFFICIAL = "https://github.com/JuliaRegistries/General.git"
const REG_URL_FORK = "https://edls-group:$(ENV["EDLS_GITHUB_TOKEN"])@github.com/edls-group/General.git"


## 1. parse package info

import TOML, LibGit2

@info "Parsing Project.toml"
project=TOML.parsefile("Project.toml")
@info "Found $(project["name"]) at v$(project["version"])"

repo = LibGit2.GitRepo(pwd())



## 2. abort if version already tagged

if "v"*project["version"] ∈ LibGit2.tag_list(repo)
    @error("A tag for version $(project["version"]) already exists in $(project["name"])")
    exit(0)
end



## 3. Set up & check General registry

import Pkg


# add General registry if it doesn't exist already
if isempty(Pkg.Registry.reachable_registries())
    @info "General registry not set up, adding manually "
    Pkg.Registry.add(Pkg.RegistrySpec(url=REG_URL_OFFICIAL))
else   
    @info "Updating registries"
    Pkg.Registry.update()
end
reg = only(Pkg.Registry.reachable_registries())

@assert reg.repo == REG_URL_OFFICIAL "Registry URL mismatch (expected $REG_URL_OFFICIAL, was $(reg.repo))"

# check whether the package is already known
uuid_in_registry = haskey(reg.pkgs, Base.UUID(project["uuid"]))
name_in_registry = !isempty(Pkg.Registry.uuids_from_name(reg, project["name"]))

# catch problematic cases
if uuid_in_registry && !name_in_registry
    @error "UUID $(project["uuid"]) already exists in the registry (package $(reg.pkgs[Base.UUID(project["uuid"])])), name changes are not possible"
    exit(301)
elseif name_in_registry && !uuid_in_registry
    @error "A package/packages named $(project["name"]) already exists in the registry (UUID $(uuids_from_name(reg, project["name"]))), name conflicts are not allowed"
    exit(302)
end

# true if this is a new package
newpackage = !(name_in_registry && uuid_in_registry)
@info "Package $(project["name"]) $(newpackage ? "doesn't exist" : "exists") in registry, preparing registration"

# registrator-style commit message
commitmsg = "New $(newpackage ? "package" : "version"): $(project["name"]) v$(project["version"])"



## 4. Initialize git repos, update fork

reg_repo = LibGit2.GitRepo(reg.path)

# set config for registration commit
reg_conf = LibGit2.GitConfig(reg_repo)
LibGit2.set!(reg_conf, "user.name", "edls-group")
LibGit2.set!(reg_conf, "user.email", "edls@ds.mpg.de")

# set up two remotes: origin (pointing at our fork)
origin = LibGit2.lookup_remote(reg_repo, "origin")
if LibGit2.url(origin) != REG_URL_FORK
    @info "Setting registry origin URL to eDLS fork"
    LibGit2.set_remote_url(reg_repo, "origin", REG_URL_FORK)
    origin = LibGit2.lookup_remote(reg_repo, "origin")
    fhead = LibGit2.fetch(origin, ["master"])
end
# and upstream (pointing at the official registry)
#upstream = LibGit2.lookup_remote(reg_repo, "upstream")
#if isnothing(upstream) 
#    @info "adding new remote "upstream" with offical repo"
#    (upstream = LibGit2.GitRemote(reg_repo, "upstream", REG_URL_OFFICIAL))
#end

# sync fork repo for force push
#LibGit2.merge!(reg_repo, branch="refs/remotes/upstream/master", fastforward=true)

# avoid unnecessary registry updates by Pkg since we have just updated the registry
Pkg.UPDATED_REGISTRY_THIS_SESSION[] = true

# branch out for testing
push_refs = ["+refs/heads/master"]
if TESTING 
    LibGit2.branch!(reg_repo, "testing", set_head = true, track = "/refs/remotes/origin/testing")
    push!(push_refs, "+refs/heads/testing")
end


## 5. Register

# install dependecies
Pkg.add("LocalRegistry")
Pkg.add("HTTP")
Pkg.add("JSON3")
import LocalRegistry, HTTP, JSON3

# acquire repository information from GitLab
gitlab_info = JSON3.read(HTTP.get(ENV["CI_API_V4_URL"]*"/projects/"*ENV["CI_PROJECT_ID"]).body)

# register, commit, push
LocalRegistry.register(pwd(), registry = reg.path, repo = gitlab_info.http_url_to_repo, push = false)
#commithash = LibGit2.commit(reg_repo, commitmsg)
# the + in front of the refspec makes it a force push
LibGit2.push(origin, push_refs)

# open PR
pr_body = JSON3.write(Dict([
    "title"=>commitmsg,
    "body"=>"""
 - Registering package: $(project["name"])
 - Repository: $(gitlab_info.http_url_to_repo)
 - Version: $(project["version"])
 - Commit: $(LibGit2.GitHash(repo, "HEAD"))
 - Description: $(gitlab_info.description)

 *This PR was created using LocalRegistry.jl and a hacky GitLab CI script. If nothing went wrong, it should be similar to a Registrator-generated PR.
 I am a machine user 🤖 and I'm currently controlled by @lhupe @philbit and @JonasIsensee.*
    """,
    "head"=>TESTING ? "testing" : "edls-group:master",
    "base"=>"master"
]))

pr_repo = TESTING ? REG_URL_FORK : REG_URL_OFFICIAL
pr_api_url = joinpath("https://api.github.com/repos", splitpath(splitext(pr_repo)[1])[3:end]..., "pulls")
@info "Making Github API call at \n    $pr_api_url\n with JSON payload "*sprint(JSON3.pretty, pr_body)

response = JSON3.read(HTTP.post(pr_api_url, ["Authorization"=>"Bearer "*ENV["EDLS_GITHUB_TOKEN"], "Accept"=>"application/vnd.github+json"], pr_body).body)

@info "Opened pull request at $(response.html_url).\n Remember to tag a release as soon as the PR is merged."

# this will be useful if we add a tag/release stage to this pipeline
# because new versions should only be tagged once the PR is merged
open(".ci/github_pr_url", "w") do io
    write(io, response.url)
end

