using Pkg
Pkg.activate(@__DIR__)

# this loads the basic InPartS package…
using InPartS
# …and this exports all the internal names to make model development easier
using InPartS.InPartSDev

using StaticArrays

# a bunch of useful Julia stdlib packages
using LinearAlgebra, Random



## 1. Model definition
#=
  In InPartS, a model is represented by objects of two types:
      - particles (<:AbstractParticle) contain the dynamic variables, forces, and particle-specific properties
      - parameters (<:AbstractParams) contain static, shared parameters
  Every particle has an associated parameters struct. Multiple particles can share the same parameters.
=#

# parameters
struct MyParticleParams <: AbstractParams
    mass::Float64
    radius::Float64
    propulsion::Float64
    rotdiffusion::Float64
    fluidviscosity::Float64
    temperature::Float64
end

# all InPartS objects need to have a kwarg-only constructor
MyParticleParams(;mass = 1.0, radius = 0.5, propulsion = 1.0, rotdiffusion = 1.0, fluidviscosity = 1.0, temperature = 1.0) = MyParticleParams(mass, radius, propulsion, rotdiffusion, fluidviscosity, temperature)


# particles
# we can automatically create the kwarg-only constructor using Base.@kwdef
Base.@kwdef mutable struct MyParticle <: AbstractParticle{2}
    # these three fields are mandatory
    id::UInt64
    centerpos::SVector{2, Float64}
    params::MyParticleParams

    velocity::SVector{2, Float64} = SA[0.0, 0.0] # default values for kwarg-constructor
    force::SVector{2, Float64} = SA[0.0, 0.0]
    orientation::Float64 = 0.0

    opacity::Float64 = 0.1
end

# this tells InPartS what kind of parameters this particle type uses
InPartS.ParamType(::Type{MyParticle}) = MyParticleParams

# this tells InPartS the maximum range of interactions caused by this particle
InPartS.interactionrange(::Type{MyParticle}, sim::Simulation) = 2maximum(p->p.radius, sim.params)

## 2. Interactions

# Lennard-Jones force shifted and rescaled to get F = 0 at r = 1
ljf(r) = -12*(2^(1/6) * r)^-13 + 6*(2^(1/6) * r)^-7

# particle-particle interaction
function InPartS.externalforces!(p1::MyParticle , p2::MyParticle, s::Simulation)
    # taking a difference vector modulo the `Domain` will automatically take periodic boundaries into account
    δ = (p1.centerpos - p2.centerpos)%s.domain

    nδ = norm(δ)
    Θ = p1.params.radius + p2.params.radius
    if nδ < Θ
        force = 50*ljf(nδ/Θ) * δ/nδ
        p1.force -= force
        p2.force += force
    end
end

# active propulsion and Stokesian drag
function InPartS.internalforces!(p::MyParticle , ::Simulation)
    activeforce = p.params.propulsion * SVector(reverse(sincos(p.orientation)))
    dragforce = 6π*p.params.fluidviscosity*p.params.radius*p.velocity
    p.force += activeforce - dragforce
end



## 3. Integration

function InPartS.step!(p::MyParticle , dt, sim::Simulation)
    # rotational diffusion
    p.orientation += p.params.temperature*p.params.rotdiffusion * √(dt) * randn(sim.rng)

    # forces and velocity diffusion
    D = params.temperature/(6π*p.params.fluidviscosity*params.radius)
    p.velocity += dt .* (p.force/p.params.mass) +  D*√(dt)*randn(sim.rng, SVector{2, Float64})

    # position change
    p.centerpos += dt .* p.velocity
end


function InPartS.reset!(p::MyParticle, ::Simulation)
    p.force = SA[0.0, 0.0]
end


## 3. Export functions
#=
  One of the key features of InPartS are the semi-automatically generated export functions, which
  can squeeze a struct into a hierarchical dictionary-like structure.
  This is great for portable output files (e.g. using JLD2)

  Export functions for a Type T are generated with the command `@genexport T`.
  We can use the `@export…` macros (see relevant docstrings) to modify the behaviour of these functions.
  Note that InPartS expects `@genexport` and the `@export…` macros to be define in the same scope
  as the type T, which is important when trying to reconstruct the data.
=#

# load JLD2 to provide an export backend
using JLD2

# disable export for the `force` field
@exportoff MyParticle force
# transform the `opacity` field for export (for no particular reason)
@exporttransform MyParticle opacity value data #((1 + value)^2) (sqrt(data) - 1)

# generate export functions
@genexport MyParticleParams
@genexport MyParticle

## 4. Visualisation
#=
  InPartS supports visualisation using PyPlot.jl. The relevant code is automatically loaded by InPartS
  when the PyPlot package is loaded.
=##

using PyPlot
pygui(true)

# This function uses PyPlot primitives to visualise a vector of MyParticle.
function PyPlot.plot(::IsPyPlottable, ps::AbstractVector{<:MyParticle};
    ax = PyPlot.gca(), offset = SA[0.0, 0.0], ec = "k", kwargs...)
    circles = map(ps) do p
        PyPlot.matplotlib.patches.Circle(p.centerpos + offset, p.params.radius)
    end
    pc = PyPlot.matplotlib.collections.PatchCollection(circles; kwargs...)
    ax.add_collection(pc)

    if offset === SA[0.0, 0.0]
        X = map(x->x.centerpos[1], ps)
        Y = map(x->x.centerpos[2], ps)
        U = map(x->cos(x.orientation), ps)
        V = map(x->sin(x.orientation), ps)
        ax.quiver(X, Y, U, V)
    end
    #ax.scatter(X, Y)

end

## 5. Simulation


sim = Simulation(MyParticle,
    domainsize = (20,20),
    boundaries = (PeriodicBoundary, PeriodicBoundary),
    rng = Random.Xoshiro(42)
)

# We now add a grid of particles to the simulation. Instead of doing this by hand, we can also use the `autogrid!` function
params = MyParticleParams(propulsion = 5.0, temperature = 1.0, rotdiffusion = 0.5)
[addparticle!(sim, MyParticle, centerpos = SA[i, j], params = params) for i ∈ 2.0:2.0:18.0, j ∈ 1.0:2.0:19.0]

# In order to observe the simulation while it is running, we add a callback for live visualisation.
# This produces a plot of the system every 0.2 simulation time units
plotcb = PeriodicCallback(InPartS.callback_plot(), 0.2)

# We also add an additional callback for exporting simulation snapshots to JLD2
savecb = InPartS.SaveCallback(JLD2.JLDFile, joinpath(@__DIR__, "demo.jld2"), interval = 1.0)


# We now evolve the system for 40 time units, using a fixed time step of 0.001 time units and the previously defined callback
evolve!(sim, 40.0, tss = FixedStepper(0.001), callback = plotcb ∘ savecb)
