using Printf
import Base.∘

export EveryStepCallback, TimedCallback, PeriodicCallback, RealTimeCallback, BurstModeCallback, CallbackSet


abstract type AbstractCallback end

"""
    prepropagate(cb::AbstractCallback, sim)
First callback hook in the main loop.
Called after force computation, but before `dt` is set and [`propagate`](@ref) is called.
The simulation state at this point is likely to be consistent (i.e. the particle configuration
is "physical" and the forces match the configuration), but your mileage may vary depending
on the model.
Use this hook to save/plot/… your simulation.
"""
prepropagate(cb::AbstractCallback, sim) = cb(sim)

"""
    postpropagate(cb::AbstractCallback, sim)
Second callback hook in the main loop
Called after `propagate` returns and the simulation time is update, as the last step in
the main loop.
"""
function postpropagate(::AbstractCallback, sim) end

"""
    finalize(cb <: AbstractCallback)
Define for custom callback if it needs to be finalized after the end of a
simulation.
"""
function finalize(::AbstractCallback, sim; kwargs...) end



"""
    NoCallback() <: AbstractCallback
Dummy callback that is used when no real callback was passed.
"""
struct NoCallback <: AbstractCallback
end
(cb::NoCallback)(args...) = true
function finalize(::NoCallback, args...; kwargs...) end


struct EveryStepCallback{F1, F2} <: AbstractCallback
    f::F1
    finalize::F2
    EveryStepCallback(f::F1, finalize::F2 = (args...; kwargs...) -> nothing) where{F1, F2} = new{F1, F2}(f, finalize)
end


(cb::EveryStepCallback)(sim) = cb.f(sim)

finalize(cb::EveryStepCallback, sim; kwargs...) = cb.finalize(sim; kwargs...)


"""
    TimedCallback(affect!, executiontimes; offset=0.0, initialize, finalize) <: AbstractCallback
Returns a callback that executes `affect!(sim)` as soon as the simulation time is larger than
the time points specified in `executiontimes`, whose elements must be accessible by index and must
be sorted in ascending order (multiple identical elements are fine).
`offset` optionally shifts all execution times by the specified amount, `initialize` and `finalize`
functions are called when the callback is first executed and at the end of `evolve`, respectively.
"""
mutable struct TimedCallback{T, U, F1, F2, F3} <: AbstractCallback
    offset::T
    executiontimes::U
    counter::Int
    initialize::F1
    affect::F2
    finalize::F3

    function TimedCallback(f::F2,  executiontimes::U; offset::T=zero(eltype(U)), initialize::F1 = (sim) -> nothing, finalize::F3 = (sim; kwargs...) -> nothing) where {T <:Number, U, F1, F2, F3}
        new{T, U, F1, F2, F3}(offset, executiontimes, -1, initialize, f, finalize)
    end
end

function (cb::TimedCallback)(sim)
    if cb.counter < 0
        cb.initialize(sim)
        cb.counter = 1
    end

    retcode::Bool = true
    while retcode && cb.counter ≤ length(cb.executiontimes) && (current_time(sim) - cb.offset) > cb.executiontimes[cb.counter]
        retcode = retcode & (cb.affect(sim) != false)
        cb.counter += 1
    end
    return retcode
end

finalize(cb::TimedCallback, sim; kwargs...) = cb.finalize(sim; kwargs...)



"""
    PeriodicCallback(affect!, interval; offset=0.0) <: AbstractCallback
Returns a callback that executes `affect!` every interval time units.
"""
mutable struct PeriodicCallback{T, F1, F2, F3} <: AbstractCallback
    offset::T
    interval::T
    counter::Int
    initialize::F1
    affect::F2
    finalize::F3

    function PeriodicCallback(f::F2,  interval::T; offset=zero(T), initialize::F1 = (sim) -> nothing, finalize::F3 = (sim; kwargs...) -> nothing) where {T <:Number,F1, F2, F3}
        new{T, F1, F2, F3}(offset, interval, -1, initialize, f, finalize)
    end
end

@deprecate PeriodicCallback(interval::Number, f; kwargs...) PeriodicCallback(f, interval; kwargs...)

function (cb::PeriodicCallback)(sim)
    if cb.counter < 0
        cb.initialize(sim)
        cb.counter = max(floor((current_time(sim) - cb.offset)/cb.interval), 0)
    end

    if floor((current_time(sim) - cb.offset)/cb.interval) ≤ cb.counter
        return
    end
    cb.counter += 1
    cb.affect(sim)
end

finalize(cb::PeriodicCallback, sim; kwargs...) = cb.finalize(sim; kwargs...)



"""
            RealTimeCallback(interval, affect!; offset=0.0) <: AbstractCallback
        Returns a callback that attempts to execute `affect!` every `interval` seconds with an initial waiting time of `offset`.
    """
mutable struct RealTimeCallback{F1, F2, F3} <: AbstractCallback
    offset::Float64
    interval::Float64
    counter::Int
    initialize::F1
    affect::F2
    finalize::F3

    function RealTimeCallback(f::F2, interval::Real; offset=time(), initialize::F1 = (sim) -> nothing, finalize::F3 = (sim; kwargs...) -> nothing) where {F1, F2, F3}
        new{F1, F2, F3}(offset, Float64(interval), -1, initialize, f, finalize)
    end
end

@deprecate RealTimeCallback(interval::Number, f; kwargs...) RealTimeCallback(f, interval; kwargs...)


function (cb::RealTimeCallback)(sim)
    if cb.counter < 0
        cb.initialize(sim)
        cb.counter = 0
    end

    triggertime = time()
    if (triggertime - cb.offset) < cb.interval
        return
    end

    cb.counter += 1
    cb.offset = triggertime
    cb.affect(sim)
end

finalize(cb::RealTimeCallback, sim; kwargs...) = cb.finalize(sim; kwargs...)



"""
    CallbackSet(args...) <: AbstractCallback
Returns a callback that combines all callbacks passed as arguments
and, if the conditions are met, executes them in that order.
"""
struct CallbackSet{T} <: AbstractCallback
    callbacks::T
end
CallbackSet(callbacks::Vararg{AbstractCallback}) = CallbackSet(callbacks)

# syntactic sugar — make sure order of execution is correct (i.e. last callback is executed last)
∘(cb1::AbstractCallback, cb2::AbstractCallback, cbs::Vararg{AbstractCallback}) = CallbackSet(reverse(cbs)..., cb2, cb1)

function prepropagate(cbs::CallbackSet, sim)
    retcode::Bool = true
    for cb in cbs.callbacks
        ret = prepropagate(cb, sim)
        retcode = retcode & (ret != false)
    end
    return retcode
end

function postpropagate(cbs::CallbackSet, sim)
    retcode::Bool = true
    for cb in cbs.callbacks
        ret = postpropagate(cb, sim)
        retcode = retcode & (ret != false)
    end
    return retcode
end


function finalize(cbs::CallbackSet, sim; kwargs...)
    for cb in cbs.callbacks
        finalize(cb, sim; kwargs...)
    end
    return
end





"""
    BurstModeCallback(; interval, affect, maxcount=10, trigger_offset=0.0, trigger_interval=10.0) <: AbstractCallback
Returns a callback that executes `affect!` every interval time units for a total of
`maxcount` iterations after it has been triggered by a `PeriodicCallback` that is
initialized with `trigger_offset` and `trigger_interval`.

## Example
A callback created with

    BurstModeCallback(; interval=0.1,
                            affect=printcurrenttime,
                            maxcount=3,)

where `printcurrenttime` does what you think it does, will, during a simulation,
print `0.0, 0.1, 0.2, 10.0, 10.1, 10.2, 20.0, ...`
"""
mutable struct BurstModeCallback{T, F1, F2, F3, CB} <: AbstractCallback
    counter::Int
    maxcount::Int
    interval::T
    offset::T
    trigger::CB
    initialize::F1
    affect::F2
    finalize::F3
end



function BurstModeCallback(
    affect::F2,
    interval::T,
    trigger::CB;
    maxcount::Int = 10,
    initialize::F1 = (sim) -> nothing,
    finalize::F3 = (sim;kwargs...) -> nothing) where {T, F1, F2, F3, CB<:AbstractCallback}

    BurstModeCallback{T, F1, F2, F3, CB}(
        -2, maxcount, interval, zero(T), trigger, initialize, affect, finalize)
end

@deprecate BurstModeCallback(interval::Number, affect, trigger::AbstractCallback; kwargs...) BurstModeCallback(affect, interval, trigger; kwargs...)

function BurstModeCallback(;
    interval,
    affect,
    maxcount = 10,
    initialize = (sim) -> nothing,
    finalize = (sim;kwargs...) -> nothing,
    trigger_offset = 0.0,
    trigger_interval = 10.0,)

    BurstModeCallback(
        affect, interval,
        PeriodicCallback(trigger_interval, (sim)->true; offset=trigger_offset);
        maxcount=maxcount,
        initialize=initialize,
        finalize=finalize)
end



function (cb::BurstModeCallback)(sim)
    if cb.counter == -2
        cb.initialize(sim)
        cb.counter = -1
    end

    if cb.counter == -1
        if cb.trigger(sim)==true
            cb.counter = 0
            cb.offset = current_time(sim)
        end
    end

    ret = nothing
    if cb.counter > -1
        if floor((current_time(sim) - cb.offset)/cb.interval) ≤ cb.counter
            return ret
        end
        cb.counter += 1
        ret = cb.affect(sim)
    end

    if cb.counter == cb.maxcount
        cb.counter = -1
    end
    return ret
end

InPartS.finalize(cb::BurstModeCallback, sim; kwargs...) = cb.finalize(sim; kwargs...)
