#=
 This file contains the main time evolution function, which calls the functions
 specified in the AbstractParticle docs to evolve a system of particles.
 Optionally, the system is divided into boxes to reduce the amount of pair
 interactions that have to be considered.
=#

export evolve!

## TIME EVOLUTION FUNCTION #####################################################

"""
    evolve!(sim, tmax; kwargs...)
Evolve a `Simulation` for `tmax` time units, using [`computeforces`](@ref) and
[`propagate`](@ref).

Returns a named tuple containing none of your business.

## Additional arguments

 - `tss::TimeSteppingStrategy` Set the time stepping strategy to use. Defaults to a [`FixedStepper`](@ref) with `dt = 1e-3`
 - `callback::AbstractCallback` Add a callback to the simulation.
 - `dontfinalize::Bool` If true, blocks execution of the callback finalizer. Use if you want to continue simulating on the same `Simulation`
 - `savedt::Bool` Include a list of timesteps `dt` in the return parameters. Not recommended except for debugging, defaults to `false`

"""
function evolve!(
    sim::Simulation, tmax::Real;
    tss = FixedStepper(0.001),
    callback::AbstractCallback = NoCallback(),
    dontfinalize = false,
    savedt::Bool = false
)

    flushadditions!(sim)

    if isempty(sim.params)
        error("Simulations without parameter structs cannot be evolved")
    end

    initialize_pc!(sim)

    # adjust tmax to account for non-zero initial times
    tmax += current_time(sim)

    if savedt
        dtlist = Float64[]
    end

    # reset particle states and apply boundary conditions via fold_back
    # to prepare for first computeforces! (usually done in propagate!)

    for obj in alldynamicobjects(sim)
        fold_back!(obj, sim.domain)
        reset!(obj, sim)
    end

    while current_time(sim) < tmax
        computeforces!(sim)

        # pre-propagate hook
        ret = prepropagate(callback, sim)
        ret == false && break

        timestep = nextdt(tss, sim)
        propagate!(sim, timestep)

        # this should be the only time in InPartS that `sim.t` and `sim.step` are used without accessor functions
        # (except for import of course…)
        sim.t += timestep
        sim.step += 1
        savedt && push!(dtlist, timestep)

        # post-propagate hook 📬🪝
        ret = postpropagate(callback, sim)
        ret == false && break
    end

    returnparams = savedt ? (tmax = tmax, tss = tss, dtlist = dtlist) : (tmax = tmax, tss = tss)

    # post-simulation hook
    dontfinalize || finalize(callback, sim; returnparams...)

    return returnparams
end

## PROPAGATION HELPER FUNCTION #################################################

"""
    propagate!(sim, dt)
Advances the system of particles by one time step using the precomputed forces.
"""
function propagate!(sim::Simulation, dt)
    iter  = alldynamicobjects(sim)

    # TODO: make this more elegant
    if isthreaded(sim)
        Threads.@threads for n in eachindex(iter)
            particle = iter[n]
            step!(particle, dt, sim)
        end
    else
        for n in eachindex(iter)
            particle = iter[n]
            step!(particle, dt, sim)
        end
    end

    flushdeletions!(sim)
    flushadditions!(sim)


    iter2 = alldynamicobjects(sim)
    if isthreaded(sim)
        Threads.@threads for n in eachindex(iter2)
            particle = iter[n]
            fold_back!(particle, sim.domain)
            reset!(particle, sim)
        end
    else
        for n in eachindex(iter2)
            particle = iter[n]
            fold_back!(particle, sim.domain)
            reset!(particle, sim)
        end
    end

    return nothing
end

## FORCE COMPUTATION ###########################################################

"""
    computeforces!(sim)
Computes the external and internal forces in the simulation.
"""
@forcedef function computeforces!(sim::Simulation)
    if isthreaded(sim)
        @forcefun threaded_computeforces!(sim, sim.particles)
    else
        @forcefun serial_computeforces!(sim, sim.particles)
    end
end

# fallback for non-parallelised PCs
@forcedef threaded_computeforces!(sim, pc) = @forcefun serial_computeforces!(sim, pc)


"""
    computeforces!(particles, sim, ::NoBoxing)
Computes all external and internal forces of `particles` naively, computing the
interactions for all pairs of particles.
"""
@forcedef function computeforces!(particles::AbstractVector{T},
                             sim::Simulation,
                             ::NoBoxing) where {T <: AbstractParticle}
    N = length(particles)
    for i = 1:N
        c1 = particles[i]
        @forcefun internalforces!(c1, sim)
        for j = i+1:N
            @forcefun externalforces!(c1, particles[j], sim)
        end
    end
end



"""
    computeforces!(particles1, particles2, sim)
Computes the external forces arising due to interaction of particles from `particles1`
with particles from `particles2`.

The function assumes that `particles1` and `particles2` do not share elements.
This condition is only checked within an [`@assert`](@ref), so there might be no
signs of things going wrong depending on the optimization level!
"""
@forcedef function computeforces!(particles1::AbstractVector{T}, particles2::AbstractVector{T},
                             sim::Simulation) where {T <: AbstractParticle}

    for c1 ∈ particles1, c2 ∈ particles2
        @assert c1 != c2
        @forcefun externalforces!(c1, c2, sim)
    end
end
