export SaveCallback, BackupCallback, readsim

## Datafile interface definitions
"""
    dfopen(backend, filename; kwargs...)
    dfopen(backend, filename, mode::String)
Open the data file at `filename`.

Mode specifications follow [`Base.open`](@ref).
"""
dfopen(type::Type, filename::String, mode::String) = dfopen(type, filename; modetoflags(mode)...)
dfopen(@nospecialize(type::Type), filename::String; kwargs...) = error("Unknown export backend: $type")

function dfopen(filename::String, mode::String="r")
    s = FileIO.query(filename)
    modulename = typeof(s).parameters[1].parameters[1]
    if !isdefined(Main, modulename)
        error("File appears to be a $modulename file but library is not loaded.")
    end
    if modulename == :JLD2
        return dfopen(JLD2.JLDFile, filename, mode)
    elseif modulename == :HDF5
        return dfopen(H5Wrapper, filename, mode)
    else
        error("This backend is not implemented")
    end
end

function dfopen(f::Function, args...; kwargs...)
    file = dfopen(args...; kwargs...)
    try
        f(file)
    finally
        dfclose(file)
    end
end


function dfclose(file) end
function dfrename(type::Type, oldname, newname; force = false) end

function gcreate(group, key) end
function isgroup(group, key) end

# why does this not already exist‽‽‽
function modetoflags(mode::String)
    if mode == "r"
        return Base.open_flags()
    elseif mode == "w"
        return Base.open_flags(write = true)
    elseif mode == "a"
        return Base.open_flags(append = true)
    elseif mode == "r+"
        return Base.open_flags(read = true, write = true)
    elseif mode == "w+"
        return Base.open_flags(truncate = true, read = true)
    elseif mode == "a+"
        return Base.open_flags(append = true, read = true)
    else
        throw(ArgumentError("invalid open mode: $mode"))
    end
end

## Callbacks


"""
    SaveCallback{backend}(trigger::AbstractCallback, filename, dumprng; [warn = true])
Create a callback that periodically saves simulation snapshots to `filename`.

# Extended help

A snapshot is saved every time the `trigger` callback is triggered.
On first activation, static simulation parameters are saved.

If the two-argument function `dumprng`, called on the `SaveCallback` and Simulation,
returns `true`, the full RNG state is dumped to the snapshot.
"""
mutable struct SaveCallback{FT, CB<:AbstractCallback, FN<:Function} <: AbstractCallback
    trigger::CB
    filename::String
    nextsnap::UInt64 # as if anyone is going to write more than typemax(Int64) snaps…
    initialized::Bool
    dumprng::FN
    warn::Bool
    SaveCallback{FT}(trigger::CB, filename::String, fn::FN, warn::Bool) where {FT, CB, FN} =
        new{FT, CB, FN}(trigger, filename, 0, false, fn, warn)
end

"""
    SaveCallback(backend, filename; [interval = 1.0], [offset = 0.0], [dumprng = (s, sim) -> (s.nextsnap % 10 == 0)], [warn = true])
Create a [`SaveCallback`](@ref) that periodically saves simulation snapshots to `filename`.

Uses a [`PeriodicCallback`](@ref) with the given `interval` and `offset`.
"""
SaveCallback(backend, filename; interval = 1.0, offset = 0.0,  dumprng = (s, sim) -> _defaultdumprng(sim.rng, s), warn = true) =
    SaveCallback{backend}(PeriodicCallback(sim -> true, interval; offset = offset), filename, dumprng, warn)


function (s::SaveCallback)(sim::Simulation)
    if !s.initialized
        _init(s, sim)
        s.initialized = true
    end
    if !(s.trigger(sim) == true)
        return
    end

    _savesnap(s, sim)
    return
end

# Xoshiros are small and can be dumped every snapshot, MersenneTwisters are huge and shouldn't be written as often
_defaultdumprng(::Random.Xoshiro, _) = true
_defaultdumprng(::Random.MersenneTwister, s) = (s.nextsnap % 10 == 0)

function _init(s::SaveCallback{FT}, sim) where FT
    f = dfopen(FT, s.filename, "w")
    try
        writestatic(f, sim; s.warn)
        # write initial snapshot with RNG dump
        writesnap(f[SNAPGROUPNAME], sim; name = snapname!(s), dumprng = true, s.warn)
    finally
        dfclose(f)
    end
    return
end

function _savesnap(s::SaveCallback{FT}, sim; dumprng = s.dumprng(s, sim)) where FT
    f = dfopen(FT, s.filename, "a")
    try
        writesnap(f[SNAPGROUPNAME], sim; name = snapname!(s), dumprng = dumprng, s.warn)
    finally
        dfclose(f)
    end
    return
end

function finalize(s::SaveCallback{FT}, sim; kwargs...) where FT
    f = dfopen(FT, s.filename, "a")
    try
        # write final snapshot with RNG dump
        writesnap(f[SNAPGROUPNAME], sim; name = snapname!(s), dumprng = true, s.warn)
        writefinal(f, sim; s.warn)
    finally
        dfclose(f)
    end
    return
end


function snapname!(s::SaveCallback)
    sn = string(s.nextsnap)
    s.nextsnap += 1
    return sn
end



"""
    BackupCallback{backend}(trigger::AbstractCallback, filename; [warn = true])
Creates a callback that dumps the entire simulation state to `filename`.

# Extended help

The entire simulation state, consisting of static parameters and a full snapshot, is saved
every time the `trigger` callback is triggered.
The data file is truncated before writing.
"""
struct BackupCallback{FT, CB} <: AbstractCallback
    trigger::CB
    filename::String
    warn::Bool
    BackupCallback{FT}(trigger::CB, filename, warn) where {FT, CB} = new{FT, CB}(trigger, filename, warn)
end

"""
   BackupCallback(backend, filename; [interval = 300.0], [offset = 0.0], [warn = true])
Creates a [`BackupCallback`](@ref) that periodically dumps the entire simulation state to `filename`.

Uses a [`RealTimeCallback`](@ref) with the given `interval` and `offset`.
"""
BackupCallback(backend, filename; interval = 300.0, offset = 0.0, warn = true) =
    BackupCallback{backend}(RealTimeCallback(interval, sim -> true; offset = offset), filename, warn)


function(cb::BackupCallback{FT})(sim::Simulation) where FT
    if !(cb.trigger(sim) == true)
        return
    end

    tf = joinpath(dirname(cb.filename), "#"*basename(cb.filename))
    f = dfopen(FT, tf, "w")
    try
        writestatic(f, sim)
        # backup files only contain one snapshot
        writesnap(f[SNAPGROUPNAME], sim; name = "0", dumprng = true)
        writefinal(f, sim)
    finally
        dfclose(f)
    end
    # only overwrite potential existing file if saving was successful
    dfrename(FT, tf, cb.filename; force = true)
    return
end

finalize(cb::BackupCallback, sim; kwargs...) = cb(sim)




## Generic writing functions
# writestatic
"""
    writestatic(simgroup, sim; [warn = true]) → snapgroup, paramgroup
Writes the static bits of the simulation to the output group `simgroup`.

This includes the particle and parameter types, the domain and some InPartS version
information.
"""
function writestatic(simgroup, sim::Simulation{PTC, PT, PM}; warn = true) where {PTC, PT, PM}
    simgroup[TREEHASHNAME] = InPartS.TREE_HASH
    simgroup[VERSIONNAME] = string(InPartS._INPARTS_VERSION)

    writetype(simgroup, PTC, name = PTCTYPENAME)
    writetype(simgroup, PT; name = PTTYPENAME)
    writetype(simgroup, PM, name = PMTYPENAME)
    writedomain(simgroup, sim.domain; warn)

    writerng(gcreate(simgroup, RNGNAME), sim.rng; complete = true, warn)

    # Keyword argument is only used by JLD2 for performance. HDF5 ignores it.
    gcreate(simgroup, SNAPGROUPNAME; est_num_entries = EST_NUM_SNAPSHOTS)
    gcreate(simgroup, PARAMGROUPNAME)

    # write params already present at the start
    # if new params are added during the simulation, they will be added during finalization
    writeparams(simgroup[PARAMGROUPNAME], sim.params; warn)

    #obstacles & stuff go here
    if hasobstacles(PTC)
        writeobstacles(simgroup, sim)
    end
    return
end


function writerng(group, rng::T; complete::Bool = false, warn = true) where {T<:AbstractRNG}
    group["_type"] = string(T)
    group["_rng_complete"] = (T == Xoshiro) || complete # Xoshiro is always completely serialized
    if complete
        _serialize_rng_full!(group, rng; warn)
    else
        _serialize_rng_partial!(group, rng; warn)
    end
    return
end



"""
    writetype(group, PT::Type; name, [warn = true])
Creates a dataset of name `name` in `group` and saves the type name of `PT`.
If `PT` is a `Union`, the union types´ names will be written instead as a vector.
"""
function writetype(group, PT::Type; name, prefix = "", warn = true)
    if PT isa Union
        typenames = string.(Base.uniontypes(PT))
        group[name] = prefix.*typenames
    else
        group[name] = prefix*string(PT)
    end
    return
end


"""
    writedomain(group, domain; [name = DOMAINGROUPNAME], [warn = true])
Creates a new subgroup called `name` in `DOMAINSIZENAME` and fills it withthe static properties
of the domain (to wit, size & boundaries).
"""
function writedomain(group, domain::Domain2D{BCX, BCY}; name = DOMAINGROUPNAME, warn = true) where {BCX, BCY}
    wgroup = gcreate(group, name)
    wgroup[DOMAINSIZENAME] = domain.size
    wgroup[DOMAINBCNAME] = [string(BCX), string(BCY)]
    return
end

"""
    writedoamin(group, domain; [name = DOMAINGROUPNAME], [warn = true])
Creates a new subgroup called `name` in `group` and fills it withthe static properties
of the domain (to wit, size & boundaries).
"""
function writedomain(group, domain::Domain3D{BCX, BCY, BCZ}; name = DOMAINGROUPNAME, warn = true) where {BCX, BCY, BCZ}
    dgroup = gcreate(group, name)
    dgroup[DOMAINSIZENAME] = domain.size
    dgroup[DOMAINBCNAME] = [string(BCX), string(BCY), string(BCZ)]
    return
end

# writesnap (snapshots)

"""
    writesnap(group, sim; name = nextsnapname(group), [warn = true], kwargs...)
Creates a new subgroup called `name` in `group` and saves a snapshot of the current
state of the dynamical properties of `sim`.

### Optional arguments

  * `isunion` whether the particle type of the simulation is a type union. Defaults to
    `PT isa Union`, where `PT` is the particle type parameter of the simulation.
  * `dumprng` whether to save (that is, write all fields recursively) the current state
    of the random number generator. Defaults to true if the simulation does not use
    the global RNG.
"""
function writesnap(group, sim::Simulation{PTC, PT};
    name = nextsnapname(group),
    isunion = (PT isa Union),
    dumprng = !(sim.rng isa Random._GLOBAL_RNG),
    warn = true
  ) where {PTC, PT}
    snapgroup = gcreate(group, name)

    snapgroup[SNAPTIMENAME] = sim.t
    snapgroup[SNAPSTEPNAME] = sim.step
    snapgroup[SNAPRTSNAME] = time()
    snapgroup[SNAPNEXTIDNAME] = sim.registrar.next_id

    if dumprng
        writerng(gcreate(snapgroup, RNGNAME), sim.rng; complete = true, warn)
    end

    if isunion
        types, masks = InPartS.maskbytype(particles(sim))
        for (T, m) ∈ zip(types, masks)
            writepgroup(snapgroup, Vector{T}(@view(particles(sim)[m])), sim; warn)
        end
    else
        writepgroup(snapgroup, particles(sim), sim; warn)
    end
    return
end


"""
    writepgroup(group, particles::Vector{PT}, sim; [name = string(PT)], [warn = true])
Tries to serialize the `particles` into a new subgroup `name` of `group`.

Do not expect this to work if `PT` is a Union.
"""
function writepgroup(group, particles::AbstractVector{PT}, sim::Simulation; name = string(PT), warn = true) where {PT}
    pgroup = gcreate(group, name)
    ExportUtils.serialize!(pgroup, particles; sim = sim)
end

# writefinal (final things)

"""
    writefinal(group, sim; [warn = true])
Write the information that is only available at the end of the simulation.

At the moment, this includes the registrar and the parameter vector
"""
function writefinal(group, sim::Simulation; warn = true)
    writedict(group, Dict(sim.registrar); name = REGNAME)
    writeparams(group[PARAMGROUPNAME], sim.params; warn)
end


"""
    writeparams(group, params; [warn = true])
Tries to serialize the `params` into a new subgroup of `group` named i, the index
of the param struct .

In contrast to [`writepgroup`](@ref), this will actually work for abstract/union
eltypes.
"""
function writeparams(group, params::AbstractVector{PT}; warn = true) where {PT<:AbstractParams}
    for (i, p) ∈ enumerate(params)
        if string(i) ∉ keys(group)
            paramgroup = gcreate(group, string(i))
            ExportUtils.serialize!(paramgroup, p)
            writetype(paramgroup, typeof(p); name = "_type", warn)
        end
    end
end

# TODO: circular references? what circular references?
function writedict(group, dict::AbstractDict; name, warn = true)
    dgroup = gcreate(group, name)
    for (k, v) ∈ pairs(dict)
        if v isa AbstractDict
            writedict(dgroup, v; name = string(k), warn)
        else
            dgroup[string(k)] = v
        end
    end
end


## generic reading functions

## Convenience function for reading things

"""
    readsim(backend, filename; snap, [warn = true], [kwargs...])
    readsim(f; snap, [warn = true], [kwargs...])
Read a simulation from file and load the specified snapshot.

Additional keyword arguments are passed through to [`readstatic!`](@ref) and [`readsnap`](@ref).
"""
function readsim(f; snap, warn = true, kwargs...)
    sim = readstatic(f; warn, kwargs...)
    try
        readsnap!(sim, f, snap; warn, kwargs...)
    catch e
        warn && @warn "Could not read snapshot: $e\nSimulation may be invalid."
        throw(e)
    end
    return sim
end

function readsim(backend, filename; kwargs...)
    f = dfopen(backend, filename, "r")
    sim = missing
    try
        sim = readsim(f; kwargs...)
    finally
        dfclose(f)
    end
    return sim
end

# readstatic (simulation)

"""
    readstatic(simgroup; [time], [warn = true]) → sim::Simulation
Reconstructs a [`Simulation`](@ref) from the information saved in `simgroup`.

To load a specific snapshot into the `Simulation`, use [`readsnap!`](@ref).

### Optional arguments
  * `time` specify the simulation time in  time units and  steps. Defaults to `(0.0, 0)`
  * `warn` turns on various warning messages. Defaults to `true`
"""

function readstatic(simgroup; time = (0.0, 0), warn::Bool = true, version = get_version(simgroup))

    if version < v"0.2.0-alpha"
        # legacy import for InPartS v1 files (probably only works for CG models)
        @debug "Using legacy import functions for InPartS v0.1"
        return legacy_load(simgroup; time = time)
    elseif version < v"0.3.0-alpha"
        # legacy import for InPartS v2 files (type signature translator)
        @debug "Using legacy import functions for InPartS v0.2"
        return readstatic_v2(simgroup; time, warn)
    end

    dgroup = haskey(simgroup, DOMAINGROUPNAME) ? simgroup[DOMAINGROUPNAME] : simgroup["world"]
    domain = readdomain(dgroup; warn)

    pttype = readtype(simgroup[PTTYPENAME], super = AbstractParticle; warn)

    ptctype = try
        readtype(simgroup[PTCTYPENAME], super = AbstractParticleContainer; warn)
    catch e
        warn && @warn "Could not load ParticleContainer:$e\n Using SimpleParticleContainer\n"
        SimpleParticleContainer{ndims(domain), pttype}
    end

    pmtype = readtype(simgroup[PMTYPENAME], super = AbstractParams; warn)


    params = readparams(simgroup[PARAMGROUPNAME]; warn)
    if REGNAME in keys(simgroup)
        reg = AbstractRegistrar(readdict(simgroup[REGNAME]))
    else
        warn && @warn "No registrar was found in the file"
        reg = SimpleSeqRegistrar{UInt64}()
    end
    rng = readrng(simgroup[RNGNAME]; complete = true, warn)
    ptc = ptctype()
    # obstacles & stuff go here
    if hasobstacles(ptctype)
        obstacles = readobstacles(simgroup[OBSTACLEGROUPNAME]; warn)
        addobstacles!(ptc, obstacles)
    #else
    end
    sim = Simulation(ptc, pttype, pmtype, domain, time..., reg, rng)
    sim.params = params
    return sim
end


"""
    readdomain(group; [warn = true])
Reconstructs a [`Domain2D`](@ref) or [`Domain3D`](@ref) from `group`.
"""
function readdomain(group; warn = true)
    size = group[DOMAINSIZENAME]
    boundarystrings = group[DOMAINBCNAME]
    boundaries = InPartS.reconstruct_subtype.(boundarystrings, Boundary)
    T = length(boundaries) == 3 ? Domain3D : Domain2D
    return T{boundaries...}(size)
end


"""
    readtype(pt; [super], [warn = true]) → T::Type
Reconstructs a type from a name as encoded with [`writetype`](@ref)

### Optional arguments
 * `super` limits the type reconstruction to children of a supertype. Defaults to `Any`
 * `mod` module in which to look for the type name. Defaults to `Main`
"""
function readtype(pt; super = Any, mod = Main, warn = true)
    #pt = simgroup[name]
    if pt isa Vector
        return Union{InPartS.reconstruct_subtype.(pt, super; mod)...}
    else
        return InPartS.reconstruct_subtype(pt, super; mod)
    end
end


"""
    readdict(group; [maxdepth], [warn = true]) → d::Union{Missing, Dict{String, Any}}
Recursively reads `group` into a dictionary until depth `maxdepth`. If
`maxdepth` is reached, all data is replaced by `missing`

## Optional arguments
  * `maxdepth` Sets the maximum recursion depth. Defaults to 32.
"""
function readdict(group; maxdepth = 32, warn = true)
    if maxdepth == 0
        warn && @warn "readdict: Maximum recursion depth reached"
        return missing
    end

    dict = Dict{String, Any}()
    for gname ∈ keys(group)
        if isgroup(group, gname)
            dict[gname] = readdict(group; maxdepth = maxdepth - 1)
        else
            dict[gname] = group[gname]
        end
    end

    return dict
end




"""
    readrng(group; [complete], [warn = true]) → rng::AbstractRNG
    readrng(group, T ; [complete], [warn = true]) → rng::T
Reconstructs the RNG from `group`. Optionally specifiy the RNG type.

## Optional Arguments
* `complete = true` try to completely reconstruct the RNG. If this is impossible, a warning will be printed.
* `mod` module in which to look for RNG type. Defaults to `InPartS.Random`
"""
readrng(group; mod = InPartS.Random, kwargs...) = readrng(group, InPartS.reconstruct_subtype(group["_type"], AbstractRNG; mod); kwargs...)

function readrng(group, T::Type{<:AbstractRNG}; complete::Bool = false, warn = true)
    if complete
        if group["_rng_complete"]
            return _deserialize_rng_full(group, T; warn)
        else
            @warn "Attempted to do a full RNG deserialization without data, falling back to partial deserialization"
        end
    end
    return _deserialize_rng_partial(group, T; warn)
end




"""
    readparams(group; [warn = true]) → params
Reconstructs the parameter structs from `group`
"""
function readparams(group; warn = true)
    indices = Vector{Int}()
    params = Vector{AbstractParams}()
    for n ∈ keys(group)
        paramindex = parse(Int, n)
        paramtype = InPartS.reconstruct_subtype(group[n]["_type"], AbstractParams)
        push!(indices, paramindex)
        push!(params, ExportUtils.deserialize(group[n], paramtype))
    end
    sp = sortperm(indices)
    return params[sp]
end



@deprecate readsnap(snapgroup; sim::Simulation) readsnap(snapgroup, sim)

"""
    readsnap(snapgroup, sim::Simulation; [warn = true]) → particles, time, step, next_id, rng
Loads the snapshot from `snapgroup`, using additional information from `sim`
"""
function readsnap(snapgroup, sim::Simulation{PTC,PT}; warn = true) where {PTC, PT}
    time = snapgroup[SNAPTIMENAME]
    step = snapgroup[SNAPSTEPNAME]
    next_id = snapgroup[SNAPNEXTIDNAME]
    particles = PT[]

    snapnames = keys(snapgroup)
    for subname ∈ snapnames
        # bodge job
        if subname ∈ [SNAPTIMENAME, SNAPSTEPNAME, SNAPNEXTIDNAME, RNGNAME, SNAPRTSNAME]
            continue
        end
        type = InPartS.reconstruct_subtype(subname, AbstractParticle)
        append!(particles, ExportUtils.deserialize(snapgroup[subname], type; sim = sim))
    end

    sort!(particles, by=x->x.id)

    rng = missing
    if RNGNAME ∈ snapnames
        rng = readrng(snapgroup[RNGNAME]; complete = true, warn)
    end

    return particles, time, step, next_id, rng
end


"""
    readsnap!(sim::Simulation, f, snapname::String; [warn = true])
    readsnap!(sim::Simulation, f, snapid::Int; [warn = true])
    readsnap!(sim::Simulation, snapgroup; [warn = true])
Loads the snapshot data from `snapgroup` into `sim`.
"""
function readsnap!(sim::Simulation, snapgroup; warn::Bool = true)
    particles, time, step, next_id, rng = readsnap(snapgroup, sim; warn)
    setparticles!(sim.particles, particles)
    sim.t = time
    sim.step = step
    sim.registrar.next_id = next_id
    if ismissing(rng)
    	warn && @warn "Could not reconstruct rng"
    else
    	(sim.rng = rng)
    end
    return sim
end

function readsnap!(sim::Simulation, f, snap; warn::Bool=true)
    if length(keys(f)) == 1
        legacy_readsnap!(sim, f, snap; warn=warn)
    else
        readsnap!(sim, f[SNAPGROUPNAME][string(snap)]; warn)
    end
end

function numsnaps(f)
    ks = keys(f)
    if length(ks) == 1
        length(keys(f[only(ks)]["snapshots"]))
    else
        length(keys(f[SNAPGROUPNAME]))
    end
end

## Stuff

snapname(i::Int) = string(i)

function nextsnapname(g)
    @warn "You really shouldn't be using this!"
    snaps = sort!(parse.(Int, keys(g)))
    return snaps[end] + 1
end


function latest_full_snap(f)
    for n = (numsnaps(f)-1):-1:0
        if haskey(f[SNAPGROUPNAME][string(n)], "rng")
            return n
        end
    end
    return -1
end

## RNG stuff

function _serialize_rng_partial!(group, @nospecialize rng::AbstractRNG; warn = true)
    warn && @warn "Partial serialization not implemented for $(typeof(rng))"
    return
end

function _serialize_rng_full!(group, @nospecialize rng::AbstractRNG; warn = true)
    warn && @warn "Full serialization not implemented for $(typeof(rng)), falling back to partial serialization"
    return _serialize_rng_partial!(group, rng; warn)
end

function _serialize_rng_partial!(group, rng::MersenneTwister; warn = true)
    group["seed"] = rng.seed
    return
end

function _serialize_rng_full!(group, rng::MersenneTwister; warn = true)
    _serialize_rng_partial!(group, rng)
    group["vals"] = rng.vals
    group["ints"] = string.(rng.ints)
    group["state"] = rng.state.val
    group["idxF"] = rng.idxF
    group["idxI"] = rng.idxI
    return
end


function _deserialize_rng_partial(group, @nospecialize T::Type{<:AbstractRNG}; warn = true)
    warn && @warn "Partial deserialization not implemented for $(T), creating new instance"
    return T()
end
_deserialize_rng_partial(group, T::Type{Random._GLOBAL_RNG}) = Random.GLOBAL_RNG

function _deserialize_rng_full(group, @nospecialize T::Type{<:AbstractRNG}; warn = true)
    warn && @warn "Full deserialization not implemented for $(T), falling back to partial deserialization"
    return _deserialize_rng_partial(group, T)
end

_deserialize_rng_partial(group, ::Type{MersenneTwister}; warn = true) = MersenneTwister(group["seed"])

function _deserialize_rng_full(group, ::Type{MersenneTwister}; warn = true)
    rng = _deserialize_rng_partial(group, MersenneTwister)

    rng.vals = group["vals"]
    rng.ints = parse.(UInt128, group["ints"])
    rng.state = Random.DSFMT.DSFMT_state(group["state"])
    rng.idxF = group["idxF"]
    rng.idxI = group["idxI"]
    return rng
end


# Xoshiro

function _serialize_rng_full!(group, rng::Xoshiro; kwargs...)
    group["s0"] = rng.s0
    group["s1"] = rng.s1
    group["s2"] = rng.s2
    group["s3"] = rng.s3
    return
end

# no reason to keep the partial serialization stuff
_serialize_rng_partial!(group, rng::Xoshiro; kwargs...) = _serialize_rng_full!(group, rng; kwargs...)

_deserialize_rng_full(group, ::Type{Xoshiro}; kwargs...) = Xoshiro(
    group["s0"], group["s1"], group["s2"], group["s3"]
)

_deserialize_rng_partial(group, T::Type{Xoshiro}; kwargs...) = _serialize_rng_full!(group, T; kwargs...)



################################################################################
#                           Obstacle Serialization                             #
################################################################################


function writeobstacles(group,  sim::Simulation; warn = true)
    sov = obstacles(sim)
    obsgroup = gcreate(group, OBSTACLEGROUPNAME)
    for (i, so) ∈ enumerate(sov)
        sogroup = gcreate(obsgroup, string(i))
        ExportUtils.serialize!(sogroup, so)
        sogroup["_type"] = string(typeof(so))
    end
end

function readobstacles(group; warn = true)
    indices = Vector{Int}()
    params = Vector{AbstractObstacle}()
    for n ∈ keys(group)
        paramindex = parse(Int, n)
        paramtype = InPartS.reconstruct_subtype(group[n]["_type"], AbstractObstacle)
        push!(indices, paramindex)
        push!(params, ExportUtils.deserialize(group[n], paramtype))
    end
    sp = sortperm(indices)
    return params[sp]
end

################################################################################
#                           Just Write A Simulation                            #
################################################################################
"""
   write_sim(filename, sim; dumprng = true)

A function for whenever you just need to save the ~~world~~ simulation.

"""
function write_sim(filename, sim::Simulation; dumprng = true)
    s = FileIO.query(filename; checkfile=true)
    modulename = typeof(s).parameters[1].parameters[1]
    if !isdefined(InPartS, modulename)
        Base.require(Main, modulename)
        return Base.invokelatest(write_sim, filename, sim; dumprng)
    elseif modulename == :JLD2
        FT = JLD2.JLDFile
    elseif modulename == :HDF5
        FT = H5Wrapper
    else
        error("This backend is not implemented")
    end
    dfopen(FT, filename, "w") do f
        writestatic(f, sim)
        writesnap(f[SNAPGROUPNAME], sim; name = "0", dumprng)
        writefinal(f, sim)
        dfclose(f)
    end
    return
end
