@forcedef function obstacleforces!(so, particles::AbstractVector{<:AbstractParticle}, sim::Simulation)
    for c in particles
        @forcefun obstacleforces!(so, c, sim)
    end
end

"""
    interactionpolygon(so, w::Domain2D; growby=0.0, kwargs...)
Return a polygon covering the obstacle `so` and additionally `growby` in all
directions beyond the boundaries of the obstacle for computing the relevant area
for interactions with particles. The `domain` may be used for proper behavior of
infinite-size obstacles.
"""
interactionpolygon(so::AbstractObstacle{2}, w; kwargs...) = getpolygon(so; kwargs...)

tostring(so::AbstractObstacle) = string(so)
###########################################################################################
#                              Axis Aligned Infinite Walls                                #
###########################################################################################
"""
    AxisAlignedWall{D, AXIS, DIRECTION}(pos) <: AbstractObstacle
An infinite wall (i.e. a (D-1)-dimensional plane) that is assumed to be axis aligned.

The type parameters determine dimensions, axis, and orientation of the wall;
the single argument gives the position along the axis of interest.

In 2D, four aliases are defined for better readability: `RightWall`, `LeftWall`,
`TopWall`, `BottomWall` with the canonical meaning.
"""
Base.@kwdef struct AxisAlignedWall{N, COMPONENT, DIRECTION} <: AbstractObstacle{N}
    val::Float64
end

@genexport(AxisAlignedWall)

normalvector(::AxisAlignedWall{2, 1, D}) where {D} = -D*SA[1.0, 0.0]
normalvector(::AxisAlignedWall{2, 2, D}) where {D} = -D*SA[0.0, 1.0]

normalvector(::AxisAlignedWall{3, 1, D}) where {D} = -D*SA[1.0, 0.0, 0.0]
normalvector(::AxisAlignedWall{3, 2, D}) where {D} = -D*SA[0.0, 1.0, 0.0]
normalvector(::AxisAlignedWall{3, 3, D}) where {D} = -D*SA[0.0, 0.0, 1.0]

const  RightWall = AxisAlignedWall{2, 1, 1}
const   LeftWall = AxisAlignedWall{2, 1,-1}
const    TopWall = AxisAlignedWall{2, 2, 1}
const BottomWall = AxisAlignedWall{2, 2,-1}



getpolygon(so::LeftWall; infval=1e3, kwargs...) =
    [so.val -infval; so.val infval; -infval infval; -infval -infval]

getpolygon(so::RightWall; infval=1e3, kwargs...) =
    [so.val -infval; so.val infval; infval infval; infval -infval]

getpolygon(so::TopWall; infval=1e3, kwargs...) =
    [-infval so.val; infval so.val; infval infval; -infval infval]

getpolygon(so::BottomWall; infval=1e3, kwargs...) =
    [-infval so.val; infval so.val; infval -infval; -infval -infval]

interactionpolygon(so::LeftWall, w::Domain2D{Boundary}; growby::Float64=0.0, kwargs...) =
    [so.val+growby 0.0; so.val+growby w.size[2]; 0.0 w.size[2]; 0.0 0.0]

interactionpolygon(so::RightWall, w::Domain2D{Boundary}; growby::Float64=0.0, kwargs...) =
    [so.val-growby 0.0; w.size[1] 0.0; w.size[1] w.size[2]; so.val-growby w.size[2]]

interactionpolygon(so::TopWall, w::Domain2D{BCX, Boundary}; growby::Float64=0.0, kwargs...) where BCX =
    [0.0 so.val-growby; w.size[1] so.val-growby;  w.size[1] w.size[2]; 0.0 w.size[2]]

interactionpolygon(so::BottomWall, w::Domain2D{BCX, Boundary}; growby::Float64=0.0, kwargs...) where BCX =
    [0.0 so.val+growby; w.size[1] so.val+growby;  w.size[1] 0.0; 0.0 0.0]


tostring(aaw::AxisAlignedWall{2, 1, 1}) = "RightWall($(aaw.val))"
tostring(aaw::AxisAlignedWall{2, 1,-1}) = "LeftWall($(aaw.val))"
tostring(aaw::AxisAlignedWall{2, 2, 1}) = "TopWall($(aaw.val))"
tostring(aaw::AxisAlignedWall{2, 2,-1}) = "BottomWall($(aaw.val))"


function relevant_boxes(so::AxisAlignedWall{3, C,D}, bg::BoxGrid, w::AbstractDomain{3}; kwargs...) where {C, D}
    intrange = minimum(bg.size)/2
    # index of the box layer one interaction range away from the wall
    extidx = unsafe_trunc(Int, (so.val - D*intrange)/(bg.size[C] + 1e-12) + 1)
    # range of affected indices in dimension C
    idxrange = D > 0 ? (extidx:size(bg.boxes)[C]) : (1:extidx)

    indices = [i for i ∈ CartesianIndices(bg.boxes) if i[C] ∈ idxrange]
end




###########################################################################################
#                                       Disk Obstacle                                     #
###########################################################################################

"""
    DiskObstacle{D}(pos, radius, inverted = false) <: AbstractObstacle
A D-dimensional spherical obstacle. With `inverted = true`, the obstacle is the entire space
*except* the D-dimensional sphere.
"""
Base.@kwdef struct DiskObstacle{D} <: AbstractObstacle{D}
    pos::SVector{D, Float64}
    radius::Float64
    inverted::Bool = false
end
# At the same time:
# Legacy positional constructor (without inverted)
# keyword version for consistency with RodObstacle
DiskObstacle{D}(pos, radius; inverted = false) where D = DiskObstacle{D}(pos, radius, inverted)

InPartS.@genexport(DiskObstacle)


function getpolygon(so::DiskObstacle{2}; res::Int=200, growby::Float64=0.0, kwargs...)
    # To make this truly a convex hull need to grow by an additional
    δ = (so.radius+growby)*(1/cos(π/res) - 1)
    radius = so.radius + growby + δ
    points = Array{Float64, 2}(undef, 2res, 2)
    for n= 1:2res
        s,c = sincos(2π * (n/2res))
        points[n, :] = so.pos + radius*SV{Float64}(c, s)
    end
    points
end



tostring(so::DiskObstacle{2}) = "DiskObstacle{2}(x=$(so.pos[1]), y=$(so.pos[2]), radius=$(so.radius), inverted=$(so.inverted))"
tostring(so::DiskObstacle{3}) = "DiskObstacle{3}(x=$(so.pos[1]), y=$(so.pos[2]), z = $(so.pos[3]), radius=$(so.radius), inverted=$(so.inverted))"

function relevant_boxes(so::DiskObstacle{3}, bg::BoxGrid, w::Domain3D{BCX, BCY, BCZ}; kwargs...) where {BCX, BCY, BCZ}
    intrange = minimum(bg.size)/2 + so.radius
    indices = CartesianIndex{3}[]
    # index of the box layer one interaction range away from the wall
    for i ∈ CartesianIndices(bg.boxes)
        boxmins = fold_back(SVector(@.((i.I - 1) * bg.size)), w)
        boxmaxs = fold_back(SVector(@.(i.I * bg.size)), w)

        cp = clamptobbox(so.pos, boxmins, boxmaxs, w)
        if norm((so.pos .- cp) % w) ≤ intrange
            push!(indices, i)
        end
    end

    indices
end


###########################################################################################
#                            Finite Wall ( Rod Obstacle )                                 #
###########################################################################################


"""
    RodObstacle(p1, p2, radius; inverted = false) <: AbstractObstacle
Construct a spherocylindrical rod between points `p1` and `p2`
with thickness `2*radius`. With `inverted = true`, the obstacle
is the entire space *except* the spherocylinder.
"""
Base.@kwdef struct RodObstacle{D} <: AbstractObstacle{D}
    backbone::LineSegment{D, Float64}
    radius::Float64
    inverted::Bool = false

    # Legacy positional constructur (without inverted)
    RodObstacle{D}(backbone, radius, inverted = false) where {D} = new(backbone, radius, inverted)
end

const FiniteWall = RodObstacle{2}

# this is a bit weird for legacy compatibility
@exporttransform RodObstacle backbone [value.p; value.eφ; value.l] begin
    dims = round(Int,(length(data) - 1)÷2)
    LineSegment{dims, Float64}(SVector{dims}(data[1:dims]), SVector{dims}(data[dims+1:2dims]), data[2dims+1])
end

InPartS.@genexport(RodObstacle)

function RodObstacle{D}(p1::SVector{D, Float64}, p2::SVector{D, Float64}, radius; inverted = false) where D
    l = norm(p2 .- p1)
    eφ = l > 0. ? SVector{D, Float64}(((p2 .- p1) ./ l)...) : SVector{D,Float64}(1.,zeros(Float64,D-1)...)
    return RodObstacle{D}(LineSegment{D, Float64}(p1, eφ, l), radius, inverted)
end



function getpolygon(so::RodObstacle{2}; res::Int= 25, growby::Float64=0.0)
    # To make this truly a convex hull need to grow by an additional
    δ = (so.radius+growby)*(1/cos(π/res) - 1)
    radius = so.radius + growby + δ
    α = π/2
    rdhat = @. radius * so.backbone.eφ#unitvector(cell.φ)

    function rRdhat(θ)
        sθ, cθ = sincos(θ)
        return SA[cθ*rdhat[1] - sθ*rdhat[2],
                  sθ*rdhat[1] + cθ*rdhat[2]]
    end

    points = Array{Float64}(undef,2res, 2)
    for (i, θ) in enumerate(range(α, stop = 2π - α, length = res))
        points[i,:] = first(so.backbone) + rRdhat(θ)
    end
    for (i, θ) in enumerate(range(α, stop = 2π - α, length = res))
        points[i + res, :] = last(so.backbone) - rRdhat(θ)
    end

    return points
end

tostring(fw::RodObstacle) = "RodObstacle(r ≈ [$(join([first(fw.backbone), last(fw.backbone)], ","))], l ≈ $(fw.backbone.l), φ ≈ $(angle(fw.backbone.eφ[1] + im*fw.backbone.eφ[2])), inverted = $(fw.inverted))"
###########################################################################################
#                                         Arc Wall                                        #
###########################################################################################

# TODO: 3D‽
"""
    ArcWall(pos, arcradius, radius, ϑmin, ϑmax) <: AbstractObstacle
Obstacle that forms an arc segment with rounded caps.
Currently 2D only.
"""
Base.@kwdef struct ArcWall <: AbstractObstacle{2}
    pos::SVector{2, Float64}
    arcradius::Float64
    radius::Float64
    ϑmin::Float64 # always -π < ϑmin < π
    ϑmax::Float64 # always ϑmin < ϑmax < 2π and ϑmax < ϑmin + 2π

    function ArcWall(pos, arcradius, radius, ϑmin, ϑmax)
        ϑmin = mod2pi(ϑmin)
        ϑmax = mod2pi(ϑmax)
        ϑmax += (ϑmax < ϑmin) ? 2π : 0
        new(pos, arcradius, radius, ϑmin, ϑmax)
    end
end

@genexport ArcWall
function getpolygon(so::ArcWall; res::Int=50, growby::Float64=0.0, kwargs...)
    # start at ϑmin in inner arc
    δϑ = (so.ϑmax - so.ϑmin) / res

    # To make this truly a convex hull need to grow by an additional
    δ = max((so.radius   +growby)*(1/cos(π/res) - 1),
            (so.arcradius+growby)*(1/cos(   δϑ) - 1))

    radius = so.radius + growby + δ
    points = Array{Float64, 2}(undef, 4res, 2)

    for n= 1:res
        ϑ = so.ϑmin + n*δϑ
        s,c = sincos(ϑ)
        points[n, :] = so.pos + (so.arcradius - radius) * SV{Float64}(c, s)
    end
    # Now do the cap
    # ϑ stays the same but φ starts at ϑ+π and goes to ϑ
    δφ = π / res
    φmax = so.ϑmax + π

    sϑ,cϑ = sincos(so.ϑmax)
    for n = 1:res
        φ = φmax - n*δφ
        sφ, cφ = sincos(φ)
        points[n+res, :] = so.pos + (so.arcradius) * SV{Float64}(cϑ, sϑ) + radius * SV{Float64}(cφ, sφ)
    end

    # Now the outer arc
    for n= 1:res
        ϑ = so.ϑmax - n*δϑ
        s,c = sincos(ϑ)
        points[n+2res, :] = so.pos + (so.arcradius + radius) * SV{Float64}(c, s)
    end

    # Now the other cap
    φmax = so.ϑmin
    sϑ,cϑ = sincos(so.ϑmin)
    for n = 1:res
        φ = φmax - δφ*n
        sφ, cφ = sincos(φ)
        points[n+3res, :] = so.pos + (so.arcradius) * SV{Float64}(cϑ, sϑ) + radius * SV{Float64}(cφ, sφ)
    end
    points
end

