#=
# Contains implementations of "Registrars" – structs for keeping track of
# "vital events" of particles. At the moment, only assigning IDs on birth is
# supported, but this could be extended in the future…
=#

export AbstractRegistrar, AbstractSeqRegistrar,
    SimpleSeqRegistrar, TrackingSeqRegistrar

"""
    AbstractRegistrar{T}
Abstract type for registrars. Subtypes should implement a function
```julia
get_id!(r::AbstractRegistrar{T}, parent::T) → T
```
that returns (preferably unique) identifers of type `T` on every call.
"""
abstract type AbstractRegistrar{T} end
# this very abstract type exists in case non-integer/non-sequential IDs
# turn out to be useful


## ABSTRACT SEQUENTIAL REGISTRAR ###############################################

"""
    AbstractSeqRegistrar{T <: Integer}
Subtype for [`AbstractRegistrar`](@ref)s that use sequential integer IDs.
Comes with an implementation of [`get_id!`](@ref) that increase the
internal counter (called `next_id`) by one.

For implementations, see [`SimpleSeqRegistrar`](@ref) or [`TrackingSeqRegistrar`](@ref).
"""
abstract type AbstractSeqRegistrar{T <: Integer} <: AbstractRegistrar{T} end


"""
    get_id!(r::AbstractSeqRegistrar{T}) → id::T
Returns the next ID. Warns on counter overflow.
"""
@inline function get_id!(r::AbstractSeqRegistrar{T}) where {T<:Integer}
    ret = r.next_id
    r.next_id += one(T)

    (r.next_id < ret) && @warn "Counter overflow! Previously assigned IDs will be reassigned!"

    return ret
end


## SIMPLE SEQUENTIAL REGISTRAR #################################################

"""
    SimpleSeqRegistrar{T<:Integer}(; next_id = )
A simple registrar that returns sequential numbers of type `T`
(defaults to `UInt64`) and does nothing else.
"""
mutable struct SimpleSeqRegistrar{T<:Integer} <: AbstractSeqRegistrar{T}
    next_id::T

    SimpleSeqRegistrar{T}(;next_id=one(T)) where {T<:Integer} = new{T}(next_id)
    SimpleSeqRegistrar(next_id::T) where {T<:Integer} = new{T}(next_id)
end

function SimpleSeqRegistrar(;next_id=nothing)
    if isnothing(next_id)
        return SimpleSeqRegistrar{UInt64}()
    else
        return SimpleSeqRegistrar(next_id)
    end
end

get_id!(reg::SimpleSeqRegistrar{T}, parent::T) where {T<:Integer} = get_id!(reg)

## TRACKING SEQUENTIAL REGISTRAR ###############################################

"""
    TrackingSeqRegistrar{T<:Integer} <: AbstractSeqRegistrar{T}
In addition to returning sequential numbers of type `T` (defaults to `UInt64`),
this implementation of [`AbstractSeqRegistrar`] also can keep track of lineages
using a dictionary called `parent`
Calling [`set_parent!`](@ref) will create a dictionary entry `(id -> parent)`.

If [`get_id!(r::TrackingSeqRegistrar{T}, parent::T) where {T<:Integer}`](@ref)
is called with the parent ID as second positional argument, the new ID is
automatically linked to its parent.
"""
mutable struct TrackingSeqRegistrar{T<:Integer} <: AbstractSeqRegistrar{T}
    next_id::T
    parent::Dict{T, T}

    TrackingSeqRegistrar{T}() where {T<:Integer} = new{T}(one(T), Dict{T, T}())
    TrackingSeqRegistrar(next_id::T, parent::Dict{T,T}) where {T<:Integer} = new{T}(next_id, parent)
end

TrackingSeqRegistrar{T}(;kwargs...) where T = TrackingSeqRegistrar(; kwargs...)
function TrackingSeqRegistrar(;next_id=nothing, parent=nothing)
    if isnothing(next_id) && isnothing(parent)
        return TrackingSeqRegistrar{UInt64}()
    end
    if isnothing(parent)
        parent = Dict{typeof(next_id),typeof(next_id)}
    end
    return TrackingSeqRegistrar(next_id,parent)
end

"""
    set_parent!(r::TrackingSeqRegistrar, id::T, parent::T)
Create an entry linking `id` to `parent` in the [`TrackingSeqRegistrar`](@ref)'s
parent list.
"""
@inline set_parent!(r::TrackingSeqRegistrar{T}, id::T, parent::T) where {T<:Integer} =
    r.parent[id] = parent

"""
    get_id!(r::TrackingSeqRegistrar{T}, parent::T) → id::T
Returns the next ID and automatically links it to the given parent ID (see
[`set_parent!`](@ref)).
"""
@inline function get_id!(r::TrackingSeqRegistrar{T}, parent::T) where {T<:Integer}
    ret = get_id!(r)
    set_parent!(r, ret, parent)
    return ret
end

function oldestancestor(id::T, reg::TrackingSeqRegistrar{T}; maxdepth = Inf) where {T<:Integer}
    current = id
    depth = 0
    while depth < maxdepth
        nextid = get(reg.parent, current, missing)
        (iszero(nextid) || ismissing(nextid)) && break
        current = nextid
        depth += 1
    end
    return current
end



makesimpletype(x) = x
makesimpletype(x::Dict{<:String,<:Any}) = x
function makesimpletype(x::Dict{T,T}) where {T<:Integer}
    m = zeros(T, length(x), 2)
    for (n, p) in enumerate(x)
        m[n, 1] = p[1]
        m[n, 2] = p[2]
    end
    m
end

makeoriginaltype(x) = x
makeoriginaltype(x::Array{T,1}) where T<:Integer = Dict{T,T}()
makeoriginaltype(x::Array{<:Integer,2}) = Dict(x[i,1] => x[i,2] for i ∈ 1:size(x)[1])

"""
    Dict(reg<:AbstractRegistrar) → dict::Dict{String, Any}
Converts a registrar object to a dictionary associating the field names with
their values.
The type of the registrar is saved using the key `"registrarType"`.

Dictionaries created using this function can be converted back to a registrar
using [`AbstractRegistrar(::Dict{String, Any}`](@ref).
"""
function Base.Dict(reg::R) where {R <: InPartS.AbstractRegistrar}
    d = Dict{String, Any}(string(fn) => makesimpletype(getproperty(reg, fn)) for fn ∈ fieldnames(R))
    d["registrarType"] = string(R)
    return d
end


"""
    AbstractRegistrar(dict::Dict{String, Any}) → concrete subtype of AbstractRegistrar
Creates a new registrar from a dictionary. This function can be used for all
subtypes of `AbstractRegistrar` that provide a constructor accepting all fields as
keyword arguments.

The type of the registrar is parsed from the value of `dict["registrarType"]`.
 Note that type parameters specified in this string are discarded. They should
therefore be inferrable from the other field values.

"""
function AbstractRegistrar(dict::Dict{String, <:Any})
    # make copy of dictionary
    dcopy = Dict(Symbol(k) => makeoriginaltype(v) for (k,v) in dict)
    # get type
    type = reconstruct_subtype(pop!(dcopy, :registrarType), InPartS.AbstractRegistrar)

    return type(;dcopy...)
end
